@@ -140,21 +140,16 @@ class NDTiffDataset():
         remote_storage_monitor : JavaObjectShadow
             Object that allows callbacks from remote NDTiffStorage. User's need not call this directly
         """
-        self._tile_width = None
-        self._tile_height = None
+        # if it is in fact a pyramid, the parent class will handle this. I think this implies that
+        # resolution levels cannot be opened seperately and expected to stitch correctly when there
+        # is tile overlap
+        self._full_resolution = False
         self._lock = threading.RLock()
         if remote_storage_monitor is not None:
             # this dataset is a view of an active acquisition. The storage exists on the java side
             self._new_image_arrived = False # used by custom (e.g. napari) viewer to check for updates. Will be reset to false by them
             self._remote_storage_monitor = remote_storage_monitor
             self.summary_metadata = self._remote_storage_monitor.get_summary_metadata()
-            if "GridPixelOverlapX" in self.summary_metadata.keys():
-                self._tile_width = (
-                    self.summary_metadata["Width"] - self.summary_metadata["GridPixelOverlapX"]
-                )
-                self._tile_height = (
-                    self.summary_metadata["Height"] - self.summary_metadata["GridPixelOverlapY"]
-                )
             if dataset_path is None:
                 self.path = remote_storage_monitor.get_disk_location()
             else:
@@ -200,8 +195,10 @@ class NDTiffDataset():
         for axes_combo in self.index.keys():
             for axis, position in axes_combo:
                 if axis not in self.axes.keys():
-                    self.axes[axis] = set()
-                self.axes[axis].add(position)
+                    self.axes[axis] = []
+                self.axes[axis].append(position)
+                # get sorted unique elements
+                self.axes[axis] = sorted(list(set(self.axes[axis])))
 
         # figure out the mapping of channel name to position by reading image metadata
         print("\rReading channel names...", end="")
@@ -218,7 +215,7 @@ class NDTiffDataset():
         print("\rDataset opened                ")
 
     def get_channel_names(self):
-        return self.channel_names.keys()
+        return list(self.channels.keys())
 
     def has_new_image(self):
         """
@@ -271,7 +268,7 @@ class NDTiffDataset():
         """
         with self._lock:
             return self._does_have_image(
-                _consolidate_axes(self.get_channel_names(), channel, channel_name, z, position, time, row, col, kwargs)
+                _consolidate_axes(self.channels, channel, channel_name, z, position, time, row, col, kwargs)
             )
 
     def read_image(
@@ -317,7 +314,7 @@ class NDTiffDataset():
 
         """
         with self._lock:
-            axes = _consolidate_axes(self.get_channel_names(),
+            axes = _consolidate_axes(self.channels,
                 channel, channel_name, z, position, time, row, col, kwargs
             )
 
@@ -363,7 +360,7 @@ class NDTiffDataset():
 
         """
         with self._lock:
-            axes = _consolidate_axes(self.get_channel_names(),
+            axes = _consolidate_axes(self.channels,
                 channel, channel_name, z, position, time, row, col, kwargs
             )
 
@@ -394,8 +391,9 @@ class NDTiffDataset():
             # update the axes that have been seen
             for axis_name in axes.keys():
                 if axis_name not in self.axes.keys():
-                    self.axes[axis_name] = set()
-                self.axes[axis_name].add(axes[axis_name])
+                    self.axes[axis_name] = []
+                self.axes[axis_name].append(axes[axis_name])
+                self.axes[axis_name] = sorted(list(set(self.axes[axis_name])))
 
             # update the map of channel names to channel indices
             self._read_channel_names()
@@ -408,20 +406,20 @@ class NDTiffDataset():
     def _read_channel_names(self):
         if 'ChNames' in self.summary_metadata:
             # It was created by a MM MDA/Clojure acquistiion engine
-            self.channel_names = {name: i for i, name in enumerate(self.summary_metadata['ChNames'])}
+            self.channels = {name: i for i, name in enumerate(self.summary_metadata['ChNames'])}
         else:
             # AcqEngJ
+            self.channels = {}
             if _CHANNEL_AXIS in self.axes.keys():
-                self.channel_names = {}
                 for key in self.index.keys():
                     axes = {axis: position for axis, position in key}
                     if (
                         _CHANNEL_AXIS in axes.keys()
-                        and axes[_CHANNEL_AXIS] not in self.channel_names.values()
+                        and axes[_CHANNEL_AXIS] not in self.channels.values()
                     ):
                         channel_name = self.read_metadata(**axes)["Channel"]
-                        self.channel_names[channel_name] = axes[_CHANNEL_AXIS]
-                    if len(self.channel_names.values()) == len(self.axes[_CHANNEL_AXIS]):
+                        self.channels[channel_name] = axes[_CHANNEL_AXIS]
+                    if len(self.channels.values()) == len(self.axes[_CHANNEL_AXIS]):
                         break
 
 
@@ -441,12 +439,9 @@ class NDTiffDataset():
 
         self.image_width = first_index["image_width"]
         self.image_height = first_index["image_height"]
-        if "GridPixelOverlapX" in self.summary_metadata:
-            self._tile_width = self.image_width - self.summary_metadata["GridPixelOverlapX"]
-            self._tile_height = self.image_height - self.summary_metadata["GridPixelOverlapY"]
 
 
-    def _add_storage_monitor_fn(self, storage_monitor_fn, callback_fn=None, debug=False):
+    def _add_storage_monitor_fn(self, acquisition, storage_monitor_fn, callback_fn=None, debug=False):
         """
         Add a callback function that gets called whenever a new image is writtern to disk (for acquisitions in
         progress only)
@@ -465,6 +460,7 @@ class NDTiffDataset():
         monitor_thread = threading.Thread(
             target=storage_monitor_fn,
             args=(
+                acquisition,
                 self,
                 push_port,
                 connected_event,
@@ -600,8 +596,13 @@ class NDTiffDataset():
         """
         if stitched and "GridPixelOverlapX" not in self.summary_metadata:
             raise Exception('This is not a stitchable dataset')
-        w = self.image_width if not stitched else self._tile_width
-        h = self.image_height if not stitched else self._tile_height
+        if not stitched or not self._full_resolution:
+            w = self.image_width
+            h = self.image_height
+        elif self._full_resolution:
+            w = self.image_width - self.overlap[1]
+            h = self.image_height - self.overlap[0]
+
         self._empty_tile = (
             np.zeros((h, w), self.dtype)
             if self.bytes_per_pixel != 3
@@ -633,6 +634,9 @@ class NDTiffDataset():
                 # get spatial layout of position indices
                 row_values = np.array(list(self.axes["row"]))
                 column_values = np.array(list(self.axes["column"]))
+                # fill in missing values
+                row_values = np.arange(np.min(row_values), np.max(row_values) + 1)
+                column_values = np.arange(np.min(column_values), np.max(column_values) + 1)
                 # make nested list of rows and columns
                 blocks = []
                 for row in row_values:
@@ -643,7 +647,9 @@ class NDTiffDataset():
                             blocks[-1].append(self._empty_tile)
                         else:
                             tile = self.read_image(**axes, **axes_to_slice, row=row, column=column, memmapped=True)
-                            if self.half_overlap[0] != 0:
+                            # remove half of the overlap around each tile so that that image stitches correctly
+                            # only need this for full resoution because downsampled ones already have the edges removed
+                            if self.half_overlap[0] != 0 and self._full_resolution:
                                 tile = tile[
                                        self.half_overlap[0]: -self.half_overlap[0],
                                        self.half_overlap[1]: -self.half_overlap[1],
@@ -651,13 +657,8 @@ class NDTiffDataset():
                             blocks[-1].append(tile)
 
                 if rgb:
-                    image = np.concatenate(
-                        [
-                            np.concatenate(row, axis=len(blocks[0][0].shape) - 2)
-                            for row in blocks
-                        ],
-                        axis=0,
-                    )
+                    image = np.concatenate([np.concatenate(row, axis=len(blocks[0][0].shape) - 2)
+                            for row in blocks],  axis=0)
                 else:
                     image = np.array(da.block(blocks))
             else:
@@ -669,8 +670,15 @@ class NDTiffDataset():
                 image = image[None]
             return image
 
+
         chunks = tuple([(1,) * len(axes_to_stack[axis]) for axis in axes_to_stack.keys()])
-        chunks += (w, h)
+        if stitched:
+            row_values = np.array(list(self.axes["row"]))
+            column_values = np.array(list(self.axes["column"]))
+            chunks += (h * (np.max(row_values) - np.min(row_values) + 1),
+                       w * (np.max(column_values) - np.min(column_values) + 1))
+        else:
+            chunks += (w, h)
         if rgb:
             chunks += (3,)
 
@@ -687,7 +695,7 @@ class NDTiffDataset():
 class NDTiffPyramidDataset():
     """Class that opens a single NDTiffStorage multi-resolution pyramid dataset"""
 
-    def __init__(self, dataset_path=None, full_res_only=True, remote_storage_monitor=None):
+    def __init__(self, dataset_path=None, remote_storage_monitor=None):
         """
         Provides access to a NDTiffStorage pyramid dataset,
         either one currently being acquired or one on disk
@@ -696,29 +704,31 @@ class NDTiffPyramidDataset():
         ----------
         dataset_path : str
             Abosolute path of top level folder of a dataset on disk
-        full_res_only : bool
-            Only open the full resolution data to save time
         remote_storage_monitor : JavaObjectShadow
             Object that allows callbacks from remote NDTiffStorage. Users need not call this directly
         """
         self._lock = threading.RLock()
         if remote_storage_monitor is not None:
-            self._remote_storage_monitor = None # IT belongs to the full resolution subclass
+            self._remote_storage_monitor = None # It belongs to the full resolution subclass
             # this dataset is a view of an active acquisition. The storage exists on the java side
             self.path = remote_storage_monitor.get_disk_location()
             self.path += "" if self.path[-1] == os.sep else os.sep
 
             with self._lock:
-                self.res_levels = {
-                    0: NDTiffDataset(remote_storage_monitor=remote_storage_monitor,
+                full_res = NDTiffDataset(remote_storage_monitor=remote_storage_monitor,
                                      dataset_path=self.path + "Full resolution" + os.sep)
-                }
+                self.res_levels = {0: full_res}
+                full_res._full_resolution = True
             # No information related higher res levels when remote storage monitoring right now
 
             #Copy stuff from the full res class for convenience
             self.summary_metadata = self.res_levels[0].summary_metadata
             self.axes = self.res_levels[0].axes
 
+            self.overlap = (np.array([self.summary_metadata["GridPixelOverlapY"],
+                                        self.summary_metadata["GridPixelOverlapX"] ]))
+            self.res_levels[0].overlap = self.overlap
+            # TODO maybe open other resoutions here too
             return
 
         # Loading from disk
@@ -738,13 +748,12 @@ class NDTiffPyramidDataset():
             )
 
         for res_dir in res_dirs:
-            if full_res_only and res_dir != "Full resolution":
-                continue
             res_dir_path = os.path.join(dataset_path, res_dir)
             res_level = NDTiffDataset(dataset_path=res_dir_path)
             if res_dir == "Full resolution":
                 with self._lock:
                     self.res_levels[0] = res_level
+                res_level._full_resolution = True
                 # get summary metadata and index tree from full resolution image
                 self.summary_metadata = res_level.summary_metadata
 
@@ -756,13 +765,14 @@ class NDTiffPyramidDataset():
                 )
 
                 self.axes = res_level.axes
-                self._channel_names = res_level.channel_names
+                self._channel_names = res_level.channels
                 self.bytes_per_pixel = res_level.bytes_per_pixel
                 self.dtype = res_level.dtype
                 self.image_width = res_level.image_width
                 self.image_height = res_level.image_height
 
             else:
+                res_level._full_resolution = False
                 with self._lock:
                     self.res_levels[int(np.log2(int(res_dir.split("x")[1])))] = res_level
 
@@ -783,7 +793,7 @@ class NDTiffPyramidDataset():
         # pass through to full resolution, since only this is monitored in current implementation
         return self.res_levels[0]
 
-    def as_array(self, axes=None, stitched=False, **kwargs):
+    def as_array(self, axes=None, stitched=False, res_level=None, **kwargs):
         """
         Read all data image data as one big Dask array with last two axes as y, x and preceeding axes depending on data.
         The dask array is made up of memory-mapped numpy arrays, so the dataset does not need to be able to fit into RAM.
@@ -800,13 +810,65 @@ class NDTiffPyramidDataset():
             The order of axes supplied in this list will be the order of the axes of the returned dask array
         stitched : bool
             Lay out adjacent tiles next to one another to form a larger image (Default value = False)
+        res_level : int or None
+            the resolution level to return. If None, return all resolutions in a list
         **kwargs :
             names and integer positions of axes on which to slice data
         Returns
         -------
         dataset : dask array
         """
-        return self.res_levels[0].as_array(axes=axes, stitched=stitched, **kwargs)
+        tile_shape = np.array([self.image_height, self.image_width]) - self.overlap
+
+
+        def get_tile_index_from_pixel_index(pixel_index):
+            """
+            :param pixel_index: pixel index at relevant resolution
+            """
+            negative_mask = pixel_index < 0
+            # positive
+            tile_index = pixel_index // tile_shape
+            # negative
+            tile_index[negative_mask] = (-1 - (np.abs(1 + pixel_index) // tile_shape))[negative_mask]
+            return tile_index
+
+
+        if res_level is not None:
+            return self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
+        else:
+            row_values = np.array(list(self.axes["row"]))
+            column_values = np.array(list(self.axes["column"]))
+            pixel_extent_min = np.array([np.min(row_values), np.min(column_values)]) * tile_shape
+            pixel_extent_max = np.array([np.max(row_values) + 1, np.max(column_values) + 1]) * tile_shape
+            images = []
+            for res_level in set(self.res_levels.keys()):
+                if res_level == 0:
+                    image = self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
+                    images.append(image)
+                else:
+                    image = self.res_levels[res_level].as_array(axes=axes, stitched=stitched, **kwargs)
+                    # crop away zero padding that extends pass where data is collected
+                    res_level_pixel_extent_min = (pixel_extent_min / 2 ** res_level).astype(np.int)
+                    res_level_pixel_extent_max = (pixel_extent_max / 2 ** res_level).astype(np.int)
+                    if np.min(np.stack([res_level_pixel_extent_max - res_level_pixel_extent_min,
+                                        res_level_pixel_extent_max - res_level_pixel_extent_min])) < 16:
+                        # Not worth it to use ones this small
+                        break
+                    # Subtract one to get pixel index from max extent because min extent is inclusive but max exclusive
+                    min_tile_index = get_tile_index_from_pixel_index(res_level_pixel_extent_min)
+                    max_tile_index = get_tile_index_from_pixel_index(res_level_pixel_extent_max - 1)
+                    # get the pixel coordinates of the tiles that contain the data
+                    res_level_container_extent_min = min_tile_index * tile_shape
+                    res_level_container_extent_max = (max_tile_index + 1) * tile_shape
+                    offset = res_level_pixel_extent_min - res_level_container_extent_min
+                    extent = res_level_pixel_extent_max - res_level_pixel_extent_min
+                    image = image[..., offset[0]: offset[0] + extent[0],
+                            offset[1]: offset[1] + extent[1]]
+                    images.append(image)
+
+
+            return images
+
 
     def has_image(
         self,
@@ -971,9 +1033,10 @@ class NDTiffPyramidDataset():
                     col=col,
                     **kwargs)
 
-    def _add_storage_monitor_fn(self, storage_monitor_fn, callback_fn=None, debug=False):
+    # TODO: this needs to be cleaned up probably--seems like this functionality belongs in pycromanager
+    def _add_storage_monitor_fn(self, acquisition, storage_monitor_fn, callback_fn=None, debug=False):
         # Only valid for the full res data
-        return self.res_levels[0]._add_storage_monitor_fn(storage_monitor_fn, callback_fn, debug)
+        return self.res_levels[0]._add_storage_monitor_fn(acquisition, storage_monitor_fn, callback_fn, debug)
 
     def _add_index_entry(self, index_entry):
         # Pass through to full res data
@@ -992,7 +1055,7 @@ _Z_AXIS = "z"
 _TIME_AXIS = "time"
 _CHANNEL_AXIS = "channel"
 
-def _consolidate_axes(channel_names, channel, channel_name, z, position, time, row, col, kwargs):
+def _consolidate_axes(channels: dict, channel: int, channel_name: str, z, position, time, row, col, kwargs):
     """
     Get all axis names in a consistent format
     """
@@ -1000,7 +1063,11 @@ def _consolidate_axes(channel_names, channel, channel_name, z, position, time, r
     if channel is not None:
         axes[_CHANNEL_AXIS] = channel
     if channel_name is not None:
-        axes[_CHANNEL_AXIS] = channel_names[channel_name]
+        if channel_name in channels.keys():
+            axes[_CHANNEL_AXIS] = channels[channel_name]
+        else:
+            raise Exception(f'{channel_name} is not a valid channel name.'
+                            f'Valid channel names are: {list(channels.keys())}')
     if z is not None:
         axes[_Z_AXIS] = z
     if position is not None:
